#ifndef FER_SOLVER_H
#define FER_SOLVER_H

/************************************************************************
   The access path represents a variable in a constraint. There are three
   types of access pathes:
  
   var       : v = var, drf = false, adr = false
   var.&     : v = var, drf = false, adr = true
   var.* + k : v = var, drf = true,  adr = false
  
   where k = 0 unless var is a function pointer, in which case k
   indicates the parameter of the target function.
 *************************************************************************/
class AccessPath {
  public:
    /* ctor: Initializes the object with default values.*/
    AccessPath(const u32 v, const u32 k, const u32 d, const u32 a);
    /* Copy constructor.*/
    AccessPath(const AccessPath& a);
    /* Pretty print.*/
    friend std::ostream & operator<<(std::ostream &, const AccessPath &);
    /* The variable represented by this object.*/
    u32 v_;
    /* The offset in case the variable is a function parameter.*/
    u16 k_;
    /* True if the access path is a dereference.*/
    bool drf_;
    /* True if the access path is an address taken.*/
    bool adr_;
};

/************************************************************************
   A constraint can is made by a pair of access paths: the left side and
   the right side. There are four basic types of constraints:
   Base: a = & b
   Simple: a = b
   Complex1: a = * b
   Complex2: * a = b
 *************************************************************************/
class Constraint {
  public:
    /* ctor: Initializes the object with default values.*/
    Constraint(AccessPath& l, AccessPath& r);
    /* Converts the internals of this object to a representation using one 64 bit value.  */
    inline const u64 to_u64() const {
      u32 a = (lhs_.drf_<<31) | (lhs_.adr_<<30) | (lhs_.k_<<23) | lhs_.v_;
      u32 b = (rhs_.drf_<<31) | (rhs_.adr_<<30) | (rhs_.k_<<23) | rhs_.v_;
      return (((u64)a << 32) | b);
    }
    /* Pretty print.*/
    friend std::ostream & operator<<(std::ostream &, const Constraint &);
    /* The left side of this constraint:*/
    AccessPath lhs_;
    /* The right side of this constraint:*/
    AccessPath rhs_;
    /* A cache used that holds the points that have been considered when
       adding edges to the constraint graph due to this constraint.
     */
    bitmap cache_;
    /* The last time when the cache of the constraint was updated.*/
    u32 last_update_;
};

/************************************************************************
   The node represents a variable in the constraint graph.
 *************************************************************************/
class Node {
  public:
    /* ctor: initializes the object with default values. It does not
       initializes the points-to-set of the object. This must be done by
       calling the method init on the object.
     */
    Node();
    /* Copy constructor.*/
    Node(const Node & n);
    /* Allocates memory for the bitmaps used by this object.*/
    void init();
    /* Frees the memory used by the bitmaps of this object.*/
    void release_memory();
    /* Pretty print.*/
    friend std::ostream & operator<<(std::ostream &, const Node &);
    /* The representative of the node. The representative will be NEG if
       the node is not part of a cycle in the constraint graph.  */
    u32 rep_;
    /* Current points-to set.*/
    bitmap pts_;
    /* Constraint edges.*/
    bitmap edges_;
    /* This field contain 8 bits that can be used to mark the node in
       different ways, dependending on the needs of the algorithm.
     */
    u8 tag_;
    /* The iteration when the node had its points-to-set updated.*/
    u32 last_update_;
    bitmap pair_;
};

class Edge {
  public:
    u32 src;
    u32 dest;
    Edge();
    Edge( u32 ,u32 );
   
};

class PairNode:public Edge{
public:
  PairNode *prev;
  PairNode *next;
  PairNode();
  PairNode(PairNode*,u32,u32,PairNode*);
};

/************************************************************************
   This class solves the inclusion based points-to-analysis using a two
   phase algorithm. In the first phase, it removes all the strongly
   connected components created in the constraint graph by simple
   constraints. After removing SCC's, the class performs a one round
   wave propagation of points-to-sets, traversing the nodes in topological
   ordering. In the second phase, the algorithm runs in iterative mode.
   At each round a complex constraint is chosen, new edges are added to
   the constraint graph, and the difference of points-to-sets is propagated
   across the constraint graph.
 *************************************************************************/
class KitSolver : public ConstraintSolver {
  public:
    /* tor.*/
    KitSolver();
    /* dtor.*/
    ~KitSolver();
    /* Solve the constraints using a two phase CDG based algorithm.*/
    void solve();
    /* Read in the constraints, create the initial constraint graph and
       points-to sets, and store the complex constraints in the
       appropriate nodes. also read in the list of variables that are
       pointer equivalent and collaps them in the constraint graph.
     */
    void init();
    /* Prints to the standard output the points-to-set of each node in the
       constraint graph.
     */
    void print_ptsto();
    /* This method computes some statistics about the algorithm and print
       the results in the standard output.
     */
    void print_statistics() ;
    /* This constraint represents an all-purpose null value. E.g, if
       get_representative(v) == ConstraintGraph::NEG, then v has no
       representative.
     */
    static const u32 NEG;
    std::list<Edge> newEdges;
    bitmap changedlist;
    u32 counter;
    PairNode **levellist;
    std::map<u32, std::map<u32,PairNode*> > ptrlist;
    void processLastLevel();
    void insertPair();
    void deletePair();
    void printLevel();
    void printEdges();
    double proc_time;//last level processing time
    double del_time;//pair deletion time
    double ins_time;//pair insertion time
    double analy_time;//total time
    double prog_time;
    double load_time;//complex1 solving time
    double store_time;//complex2 solving time
    u32 cp_constr;
    u32 adr_constr;
    u32 st_constr;
    u32 ld_constr;
    u32 K_; 
    std::string gen_file;
    std::string ptsto_file;
    std::string stats_file;
  private:
    /* These constants are used as a mask to set a bit in the tag_ field of
       a node object in order to indicate that that object is been
       visited during the deep propagation of points-to-sets. WHITE nodes
       are not been visited. GREEN nodes are been visited, and they are
       part of a cycle. RED nodes are been visited, but they do not belong
       into a cycle.
     */
    static const u8 WHITE;
    static const u8 GREEN;
    static const u8 RED;
    /* This variable is set to true when a change in the points-to-set of
       a variable is detected.
     */
    bool change_detected_;
    /* Statistics variables:  */
    u32 num_iterations_;
    /* This varible is used to identify the moment in which the
       points-to-set of a node or the cache of a constraint has been
       updated, so that only constraints with outdated caches are
       processed. */
    u32 time_stamp_;
    /* The list of nodes that constitute the constraint graph.*/
    std::vector<Node> node_list_;
    /* The list of complex constraints.*/
    std::list<Constraint> constraints_;
    /* The stack of nodes that must be collapsed into a SCC:*/
    std::stack<u32> to_be_collapsed_;
    /* The stack of nodes that are been visited during the traversal of
       the constraint graph.
     */
    std::stack<u32> visited_new_pts_;
    std::stack<u32> visited_old_pts_;
    /* The map of functions to number of parameters.*/
    std::map<u32,u32> fp_;
    /* Maximum index of the variables in the constraint graph.*/
    u32 maxv_;
    /* Minimum index of the variables in the constraint graph.*/
    u32 minv_;
    u32 maxc_;
    /* The stack of nodes that wait to fit in a SCC:*/
    std::stack<u32> stackU_;
    /* The indice of nodes in the depth-first search:*/
    std::vector<u32> dfs_;
    /* For each node v, its root is the entry node of the SCC where v
       belongs.
     */
    std::vector<u32> root_;
    /* This bitmap marks if a node has already been discovered to belong
       in a SCC component.
     */
    bitmap inComponent_;
    /* The current index of the depth-first search:*/
    u32 current_index_;
    /* The worklist that will be used in the first wave of pts propagation.
       This worklist will hold the constraint variables in topological order.
     */
    std::stack<u32> worklist_;
    /* Returns the representative of node n. This method can alter the
       contents of the constraint graph. In performs a transitive closure
       in the state of representatives, that is, if rep(v) = z, and
       rep(z) = w, then find(v) will mark rep(v) = w.
     */
    u32 find(const u32 n);
    /* Finds the representative of a node without changing the internal
       state of the object.
     */
    u32 const_find(const u32 n) const;
    /* This method updates the edges that leave from node v. It removes
       from the edge set of node v those edges that go to nodes that have
       been collapsed, and it adds to the edge set of node v the
       representatives of the collapsed nodes.
     */
    void update_edges(const u32 v);
    /* This method collapses node target into node root. After collapsing,
       find(target) == root. The points-to-set of target is merged into
       the points-to-set of root, and the edges leaving from target are
       merged into the edges leaving from root.
     */
    void unify_nodes(const u32 target, const u32 root);
    /* This method propagates the points-to-set of node v to every other
       node that is a successor of v. It does not propagate points-to-set
       to other nodes reachable from v.  */
    void shallow_propagate_pts(const u32 v);
    /* This method unify nodes that the method deep_propagate_pts has
       discovered to be in cycles.
     */
    void unify_stacked_nodes(const u32 stop_point);
    /* This method prints the points-to-set of node n, which must have
       representative rep.
     */
    void print_pts(const u32 node, const u32 rep, std::ofstream& out_) const;
    /* Marks the node v as visited during the deep propagation of a
       points-to-set. Tag must be either GREEN or RED.
     */
    inline void mark_visited_new_pts(const u32 v, const u8 tag);
    inline void mark_visited_old_pts(const u32 v, const u8 tag);
    /* Unmark all the nodes marked as visited.
     */
    inline void unmark_visited_nodes();
    /* This method solves constraint c, which must be of type complex 1, 
       e.g a = *b.
     */
    void solve_complex1(Constraint *c);
    /* This method solves constraint c, which must be of type complex 2, 
       e.g *a = b.
     */
    void solve_complex2(Constraint *c);
    /* This method propagates dif_pts to every node that is reachable from
       v. This method assumes that only paths from v to stop_point are
       cycles. If stop_point is reachable from v, than all the vertices in
       the reaching path are collapsed into a single node that has
       stop_point as the root node.
       TODO: verify if that is not copying the whole bitmap in every
       function call.
     */
    bool deep_propagate_pts(
      const u32 stop_point,
      const u32 v,
      const bitmap dif_pts
    );
    /* Finds all the strongly connected components of this graph. Unify
       nodes in the same connected component and performs one round
       propagation of points-to-sets.
     */
    void find_strongly_connected_components();
    /* This is Tarjan's strongly connected component finding algorithm, as
       modified by Nuutila to keep only non-root nodes on the stack.
       The algorithm can be found in "On finding the strongly connected
       connected components in a directed graph" by Esko Nuutila and Eljas
       Soisalon-Soininen, in Information Processing Letters volume 49,
       number 1, pages 9-14.
    */
    void visit(u32 n);
};

#endif