#include <unordered_map>
#include <map>
#include <list>
#include <stack>
#include <vector>
#include <string>
#include <sstream>
#include <fstream>
#include <iostream>
#include <sys/resource.h>
#include<sys/time.h>
#include <iomanip>
#include "debug.h"
#include "bitmap.h"
#include "obstack.h"
#include "types.h"
#include "ConstraintSolver.h"
#include "KitSolver.h"

#define LAST_LEVEL K_-1-counter
#define FIRST_LEVEL (K_-counter)%K_
#define INC_COUNTER counter=(counter+1)%K_
using namespace std;

u32 get_mem_usage(){
  unsigned x= 0;
  std::ifstream in("/proc/self/status");
  while(in){
    std::string line, tag;
    getline(in, line);
    std::istringstream iss(line);
    iss>>tag;
    if(tag == "VmData:"){
      iss>>x;
      break;
    }
  }
  in.close();
  return x;
}

double get_time() {
  struct rusage ru;
  if (getrusage(RUSAGE_SELF,&ru)) {
    return 0;
  } 
  else {
    return ru.ru_utime.tv_sec + .000001*ru.ru_utime.tv_usec + 
    ru.ru_stime.tv_sec + .000001*ru.ru_stime.tv_usec;
  }
}

const u8 KitSolver::WHITE = 0;
const u8 KitSolver::GREEN = 1;
const u8 KitSolver::RED   = 2;
const u32 KitSolver::NEG  = (u32)-1;

//=======================================================================
// AccessPath
//=======================================================================
AccessPath::AccessPath(const u32 v, const u32 k, const u32 d, const u32 a)
: v_(v), k_(k), drf_(d), adr_(a) {
}

AccessPath::AccessPath(const AccessPath & a) :
v_(a.v_), k_(a.k_), drf_(a.drf_), adr_(a.adr_) {
}

std::ostream & operator<<(std::ostream & os, const AccessPath & ap) {
  if (ap.drf_) {
    os << "*";
  }
  if (ap.adr_) {
    os << "&";
  }
  os << "var_" << ap.v_;
  if (ap.k_) {
    os << " + " << ap.k_;
  }
  return os;
}

//=======================================================================
// Constraint
//=======================================================================
Constraint::Constraint (AccessPath& l, AccessPath& r) :
lhs_(l), rhs_(r), last_update_(0) {
  cache_ = BITMAP_ALLOC(NULL);
}

std::ostream & operator<<(std::ostream & os, const Constraint & cons) {
  os << cons.lhs_ << " = " << cons.rhs_;
  return os;
}

//=======================================================================
// Node
//=======================================================================
Node::Node() :
rep_(KitSolver::NEG), pts_(0), edges_(0), tag_(0), last_update_(0) {
}

Node::Node(const Node & n) {
  rep_ = n.rep_;
  pts_ = n.pts_;
  edges_ = n.edges_;
  tag_ = n.tag_;
  last_update_ = n.last_update_;
}

void Node::init() {
  pts_ = BITMAP_ALLOC(NULL);
  edges_ = BITMAP_ALLOC(NULL);
  pair_=BITMAP_ALLOC(NULL);  
}

void Node::release_memory() {
  BITMAP_FREE(pts_);
  BITMAP_FREE(edges_);
}

std::ostream & operator<<(std::ostream & os, const Node & n) {
  os << "Representative: " << n.rep_ << std::endl;
  return os;
}

//=======================================================================
// NodePair
//=======================================================================
NodePair::NodePair() : u_(-1), v_(-1){
}

NodePair::NodePair(u32 s,u32 d) :  u_(s), v_(d){
}

//=======================================================================
// KitSolver
//=======================================================================
KitSolver::KitSolver() : change_detected_(false), num_iterations_(0),
time_stamp_(1), maxv_(0), minv_(0), current_index_(0) {
}

KitSolver::~KitSolver() {
  node_list_.clear();
}

u32 KitSolver::find(const u32 node) {
  //assert(node < node_list_.size());
  if (node_list_[node].rep_ == NEG) {
    return node;
  } 
  else {
    u32 actual_node = find(node_list_[node].rep_);
    node_list_[node].rep_ = actual_node;
    return actual_node;
  }
}

u32 KitSolver::const_find(const u32 node) const {
  //assert(node < node_list_.size());
  if (node_list_[node].rep_ == NEG) {
    return node;
  } 
  else {
    return const_find(node_list_[node].rep_);
  }
}

void KitSolver::update_edges(const u32 v) {
  //assert(v < node_list_.size());
  Node & src = node_list_[v];
  //assert(src.rep_ == KitSolver::NEG);
  /* Update the edges leaving from v. Replace the edges leading to nodes
   *    that have been collapsed by their representatives.  */
  bitmap new_edges = BITMAP_ALLOC(NULL);
  bitmap del_edges = BITMAP_ALLOC(NULL);
  u32 succ;
  bitmap_iterator bi;
  EXECUTE_IF_SET_IN_BITMAP(src.edges_, 0, succ, bi) {
    if (succ == v) {
      bitmap_set_bit(del_edges, succ);
    } 
    else if (node_list_[succ].rep_ != KitSolver::NEG) {
      u32 succ_rep = find(succ);
      if (succ_rep != v) {
	bitmap_set_bit(new_edges, succ_rep);
      }
      bitmap_set_bit(del_edges, succ);
    }
  }
  /* edges = new_edges | (edges & ~del_edges)  */
  bitmap_and_compl_into(src.edges_, del_edges);
  bitmap_ior_into(src.edges_, new_edges);
  BITMAP_FREE(new_edges);
  BITMAP_FREE(del_edges);
}

void KitSolver::shallow_propagate_pts(u32 v) {
  //assert(v < node_list_.size());
  Node & src = node_list_[v];
  //assert(src.rep_ == KitSolver::NEG);
  /* Update the edges that have changed due to collapsing nodes:  */
  update_edges(v);
  /* Propagate points-to-sets:  */
  if (!bitmap_empty_p(src.pts_)) {
    u32 succ;
    bitmap_iterator bi;
    EXECUTE_IF_SET_IN_BITMAP(src.edges_, 0, succ, bi) {
      bitmap_ior_into(node_list_[succ].pts_, src.pts_);
    }
  }
}

void KitSolver::unify_nodes(const u32 target, const u32 root) {
  PMESG("Unify (target=%d, root=%d)\n", target, root);
  Node & nr = node_list_[root];
  Node & nt = node_list_[target];
  nt.rep_ = root;
  bitmap_ior_into(nr.edges_, nt.edges_);
  bitmap_ior_into(nr.pts_, nt.pts_);
  nt.release_memory();
}

void KitSolver::unify_stacked_nodes(const u32 stop_point) {
  /* Unify stop_point with the nodes in the stack of collapsed nodes.  */
  while(!to_be_collapsed_.empty()) {
    u32 v = to_be_collapsed_.top();
    to_be_collapsed_.pop();
    if (v != stop_point && node_list_[v].rep_ == NEG) {
      unify_nodes(v, stop_point);
    }
  }
  //update_edges(stop_point);
}

inline void KitSolver::mark_visited_old_pts(const u32 v, const u8 tag) {
  node_list_[v].tag_ = tag;
  visited_old_pts_.push(v);
}

inline void KitSolver::mark_visited_new_pts(const u32 v, const u8 tag) {
  node_list_[v].tag_ = tag;
  visited_new_pts_.push(v);
}

inline void KitSolver::unmark_visited_nodes() {
  time_stamp_++;
  while(!visited_new_pts_.empty()) {
    u32 v = visited_new_pts_.top();
    visited_new_pts_.pop();
    node_list_[v].tag_ = WHITE;
    node_list_[v].last_update_ = time_stamp_;
    PMESG("v%d.lut = %d\n", v, node_list_[v].last_update_);
  }
  while(!visited_old_pts_.empty()) {
    u32 v = visited_old_pts_.top();
    visited_old_pts_.pop();
    node_list_[v].tag_ = WHITE;
  }
}

// lv = *rv
void KitSolver::solve_complex1(Constraint *c) {
  u32 rv = c->rhs_.v_;
  /* Compute the new points-to-set of node rv.  */
  if (c->last_update_ < node_list_[rv].last_update_) {
    c->last_update_ = time_stamp_++;
    bitmap new_edges = BITMAP_ALLOC(NULL);
    bitmap_and_compl(new_edges, node_list_[rv].pts_, c->cache_);
    if (!bitmap_empty_p(new_edges) ) {
      bitmap_ior_into(c->cache_, new_edges);
      /* Compute the new points-to-set of node lv.  */
      u32 lv = c->lhs_.v_;
      PMESG("Processing %d = *%d, [%d]\n", lv, rv, c->last_update_);
      bitmap new_pts = BITMAP_ALLOC(NULL);
      u32 be;
      bitmap_iterator bi;
      u32 k = c->rhs_.k_;
      EXECUTE_IF_SET_IN_BITMAP (new_edges, 0, be, bi) {
	u32 v = be;
	if(k > 0 && k > fp_[v]) {
	  continue;
	} 
	else {
	  v = find(v+k);
	}
	/* Add edge v -> lv to constraint graph:  */
	if (v != lv && bitmap_set_bit_p(node_list_[v].edges_, lv)) {
	  PMESG("New edge %d -> %d\n", v, lv);
	  newEdges.push_back(NodePair(v,lv));
	  /* Add pts(v) to the new points-to-set:  */
	  bitmap_ior_into(new_pts, node_list_[v].pts_);
	}
      }
      BITMAP_FREE(new_edges);
      /* Compute the difference that must be propagated:  */
      bitmap dif_pts = BITMAP_ALLOC(NULL);
      bitmap_and_compl(dif_pts, new_pts, node_list_[lv].pts_);
      /* Propagate the new points-to-set if there are new points:  */
      if (!bitmap_empty_p(dif_pts)) {
	if (deep_propagate_pts(lv, lv, dif_pts)) {
	  unify_stacked_nodes(lv);
	}
      }
      unmark_visited_nodes();
      BITMAP_FREE(dif_pts);
      BITMAP_FREE(new_pts);
    }
  }
}

// *lv = rv
void KitSolver::solve_complex2(Constraint *c) {
  bool there_are_nodes_to_unify = false;
  u32 lv = c->lhs_.v_;
  /* Compute the new points-to-set of node rv.  */
  if (c->last_update_ < node_list_[lv].last_update_) {
    c->last_update_ = time_stamp_++;
    bitmap new_edges = BITMAP_ALLOC(NULL);
    bitmap_and_compl(new_edges, node_list_[lv].pts_, c->cache_);
    if (!bitmap_empty_p(new_edges) ) {
      bitmap_ior_into(c->cache_, new_edges);
      u32 rv = c->rhs_.v_;
      u32 be;
      bitmap_iterator bi;
      PMESG("Processing *%d = %d, [%d]\n", lv, rv, c->last_update_);
      /* Compute the new points-to-set of the successors of node rv.  */
      u32 k = c->lhs_.k_;
      EXECUTE_IF_SET_IN_BITMAP (new_edges, 0, be, bi) {
	u32 v = be;
	if(k > 0 && k > fp_[v]) {
	  continue;
	} else {
	  v = find(v+k);
	}
	/* Add edge rv -> v to constraint graph:  */
	if (rv != v && bitmap_set_bit_p(node_list_[rv].edges_, v)) {
	  PMESG("New edge %d -> %d\n", rv, v);
	  newEdges.push_back(NodePair(rv,v));
	  if (deep_propagate_pts(rv, v, node_list_[rv].pts_)) {
	    there_are_nodes_to_unify = true;
	  }
	}
      }
      BITMAP_FREE(new_edges);
      if (there_are_nodes_to_unify) {
	unify_stacked_nodes(rv);
      }
      unmark_visited_nodes();
    }
  }
}

bool KitSolver::deep_propagate_pts(const u32 stop_point,const u32 v,const bitmap dif_pts) {
  if (node_list_[v].tag_ == GREEN) {
    return true;
  } 
  else if (node_list_[v].tag_ == RED) {
    return false;
  } 
  else if (bitmap_ior_into(node_list_[v].pts_, dif_pts)) {
    PMESG("Propagating into %d, with stop point %d\n", v, stop_point);
    bitmap_set_bit(changedlist,v);
    change_detected_ = true;
    if (bitmap_empty_p(node_list_[v].edges_)) {
      mark_visited_new_pts(v, RED);
      return false;
    }
    bool result = false;
    // Recursively propagate dif_pts to nodes reachable from succ.
    u32 succ;
    bitmap_iterator bi;
    EXECUTE_IF_SET_IN_BITMAP(node_list_[v].edges_, 0, succ, bi) {
      u32 succ_rep = find(succ);
      if ((succ_rep == stop_point) || (succ_rep != v &&
	deep_propagate_pts(stop_point, succ_rep, dif_pts))) {
	result = true;
	}
    }
    if (result) {
      if (v != stop_point) {
	to_be_collapsed_.push(v);
      }
      mark_visited_new_pts(v, GREEN);
    } 
    else {
      mark_visited_new_pts(v, RED);
    }
    return result;
  }
  else if (bitmap_bit_p(node_list_[v].edges_, stop_point)) {
    to_be_collapsed_.push(v);
    mark_visited_old_pts(v, GREEN);
    return true;
  } 
  else {
    mark_visited_old_pts(v, RED);
    return false;
  }
}

void KitSolver::find_strongly_connected_components() {
  dfs_.resize(node_list_.size(), KitSolver::NEG);
  root_.resize(node_list_.size(), KitSolver::NEG);
  inComponent_ = BITMAP_ALLOC(NULL);
  /* This first pass will mark the rep field of all the nodes in a SCC
   *    with the root of that SCC. The root node will not be marked.  */
  for (u32 i = 0; i < node_list_.size(); i++) {
    if (node_list_[i].rep_==KitSolver::NEG && dfs_[i] == KitSolver::NEG) {
      visit(i);
    }
  }
  /* This second pass unify nodes that have the rep field marked.  */
  for (u32 i = 0; i < root_.size(); i++) {
    if (root_[i] != KitSolver::NEG  && i != root_[i]) {
      unify_nodes(i, root_[i]);
    }
  }
  /* Propagate points-to-sets in topological order:  */
  while(!worklist_.empty()) {
    u32 v = worklist_.top();
    worklist_.pop();
    shallow_propagate_pts(v);
  }
  dfs_.clear();
  root_.clear();
  BITMAP_FREE(inComponent_);
}

void KitSolver::visit(u32 v) {
  dfs_[v] = current_index_ ++;
  root_[v] = v;
  /* Visit all the successors of v.  */
  bitmap_iterator bi;
  u32 w;
  EXECUTE_IF_SET_IN_BITMAP(node_list_[v].edges_, 0, w, bi) {
    if (node_list_[w].rep_ == KitSolver::NEG) {
      if (dfs_[w] == KitSolver::NEG) {
	visit(w);
      }
      if (!bitmap_bit_p(inComponent_, w)) {
	root_[v] = dfs_[root_[v]] < dfs_[root_[w]] ? root_[v] : root_[w];
      }
    }
  }
  /* The second phase of the algorithm:
   *    assign components to stacked nodes:  */
  if (root_[v] == v) {
    bitmap_set_bit(inComponent_, v);
    while (!stackU_.empty()) {
      w = stackU_.top();
      if (dfs_[w] <= dfs_[v]) {
	break;
      } 
      else {
	stackU_.pop();
	bitmap_set_bit(inComponent_, w);;
	root_[w] = v;
      }
    }
    /* Add v to the worklist in post-order of depth first search, that is,
     *      after the off-spring of v has been visited. This gives a
     *      topological ordering of the constraint graph.  */
    worklist_.push(v);;
  } 
  else {
    stackU_.push(v);
  }
}

void KitSolver::insertPair() {
  //cerr<<"start insert...."<<endl;
  u32 u,v,ur,vr;
  for(list<NodePair>::iterator it=newEdges.begin();it!=newEdges.end();it++){
    u=(*it).u_;
    v=(*it).v_;
    ur=find(u);
    vr=find(v);
    if(!(bitmap_bit_p(node_list_[u].pair_,v)) && (bitmap_equal_p(node_list_[ur].pts_,node_list_[vr].pts_))){
      bitmap_set_bit(node_list_[u].pair_,v);
      bitmap_set_bit(node_list_[v].pair_,u);
      if(v<u){
	u32 t=u;u=v;v=t;
      }
      u64 xy=u+v;
      xy=0.5*(xy*(xy+1))+v;
      NodePair np(u,v);
      levellist[FIRST_LEVEL].insert(pair<u64,NodePair>(xy,np));
    }
  }
}

void KitSolver::deletePair(){
  //cerr<<"start delete...."<<endl;
  u32 u,v,ur,vr;
  u32 be;
  bitmap_iterator bi;
  EXECUTE_IF_SET_IN_BITMAP (changedlist, 0, be, bi){
    u=be;
    ur=find(u);
    u32 be2;
    bitmap_iterator bi2;
    EXECUTE_IF_SET_IN_BITMAP (node_list_[u].pair_, 0, be2, bi2) {
      v = be2;
      vr=find(v);
      if(bitmap_bit_p(changedlist,v) && bitmap_bit_p(node_list_[u].pair_,v) && !bitmap_equal_p(node_list_[ur].pts_,node_list_[vr].pts_) ){
	bitmap_clear_bit(node_list_[u].pair_,v);
	bitmap_clear_bit(node_list_[v].pair_,u);
	u32 tu,tv;
	tu=u;
	tv=v;
	if(tv<tu){
	  u32 t=tu;tu=tv;tv=t;
	}
	u64 xy=tu+tv;
	xy=0.5*(xy*(xy+1))+tv;
	map<u64,NodePair>::iterator it;
	for(u32 i=0;i<K_;i++){
	  it=levellist[i].find(xy);
	  if(it!=levellist[i].end()){
	    levellist[i].erase(it);
	    break;
	  }
	}
      } 
    }
  }
}

void KitSolver::processLastLevel(){
  //cerr<<"start processing last level...."<<endl;
  u32 u,v,ur,vr;
  for(map<u64,NodePair>::iterator it=levellist[LAST_LEVEL].begin();it!=levellist[LAST_LEVEL].end();it++){
    u=(*it).second.u_;
    v=(*it).second.v_;
    ur=find(u);
    vr=find(v);
    if(ur!=vr){
      unify_nodes(ur,vr);
    }
    bitmap_clear_bit(node_list_[u].pair_,v);
    bitmap_clear_bit(node_list_[v].pair_,u);
  }
  levellist[LAST_LEVEL].clear();
}

void KitSolver::solve() {
  double start_time,end_time;
  // Perform the initial wave propagation.
  find_strongly_connected_components();
  // Mark the nodes that have non-empty points-to-sets
  for (u32 v = minv_; v < node_list_.size(); v++) {
    Node & node_v = node_list_[v];
    if (node_v.rep_ == KitSolver::NEG && !bitmap_empty_p(node_v.pts_)) {
      node_v.last_update_ = 1;
    }
  }
  // Solve the complex constraints:
  std::map<u64, bool> already_seen_constraints;
  do {
    num_iterations_++;
    change_detected_ = false;
    std::list<Constraint>::iterator i, e;
    for (i = constraints_.begin(), e = constraints_.end(); i != e; i++) {
      Constraint *c = &*i;
      // Check if this constraint was eliminated before.
      if (c->cache_ == 0) {
	continue;
      }
      // Update the constraint, as it may have changed due to cycle elimination.
      c->lhs_.v_ = find(c->lhs_.v_);
      c->rhs_.v_ = find(c->rhs_.v_);
      // Check if this constraint has already been processed.
      u64 c_hash = c->to_u64();
      if (already_seen_constraints[c_hash]) {
	BITMAP_FREE(c->cache_);
      } 
      else {
	already_seen_constraints[c_hash] = true;
	if (c->rhs_.drf_) {
	  // c is a complex constraint like a = *b
	  start_time=get_time();
	  solve_complex1(c);
	  end_time=get_time();
	  load_time+=(end_time-start_time);
	} 
	else {
	  // c is a complex constraint like *a = b
	  start_time=get_time();
	  solve_complex2(c);
	  end_time=get_time();
	  store_time+=(end_time-start_time);
	}
      }
    }
    already_seen_constraints.clear();
    
    start_time = get_time();
    insertPair();
    end_time = get_time();
    ins_time+=(end_time-start_time);
    
    start_time = get_time();
    deletePair(); 
    end_time = get_time();
    del_time+=(end_time-start_time);
    
    start_time = get_time();
    processLastLevel(); 
    end_time = get_time();
    proc_time+=(end_time-start_time);
    
    INC_COUNTER;
    changedlist= BITMAP_ALLOC(NULL);
    newEdges.clear();    
  } while (change_detected_);
}

void KitSolver::print_ptsto() {
  cerr<<"printing ptsto....\n";
  ofstream out_(ptsto_file.c_str());
  for (u32 i = minv_; i < maxv_; i++) {
    u32 v = find(i);
    print_pts(i, v, out_);
  }
  out_.close();
}

void KitSolver::print_pts(u32 node, u32 rep, ofstream& out_) const {
  assert(node_list_[node].rep_ == KitSolver::NEG ||node_list_[node].rep_ == rep);
  u64 n=bitmap_count_bits(node_list_[rep].pts_);
  out_<<node<<" "<<n;
  u32 be;
  bitmap_iterator bi;
  EXECUTE_IF_SET_IN_BITMAP(node_list_[rep].pts_, 0, be, bi) {
    out_ << " " << be;
  }
  out_<<"\n";
}

void KitSolver::print_statistics()  {
  cerr<<"writing stats to file....\n";
  u32 num_collapsed_nodes = 0;
  u32 num_scc = 0;
  for(u32 u = 0; u < node_list_.size(); u++) {
    if (node_list_[u].rep_ == NEG) {
      num_scc++;
    } 
    else {
      num_collapsed_nodes++;
    }
  }
  u64 tpts=0;
  for(u32 u = 0; u < maxv_; u++) {
    u32 be;
    bitmap_iterator bi;
    u32 ur=find(u);
    EXECUTE_IF_SET_IN_BITMAP (node_list_[ur].pts_, 0, be, bi){
      tpts++;
    }
  }
  u64 diffpts=0;
  for(u32 i=0;i<maxv_-1;i++){
    if(node_list_[i].pts_!=NULL){
      for(u32 j=i+1;j<maxv_;j++){
	if(node_list_[j].pts_!=NULL && bitmap_equal_p (node_list_[i].pts_,node_list_[j].pts_))
	  node_list_[j].release_memory();
      }
      diffpts++;
    }
  }
  ofstream out_;
  out_.open(stats_file.c_str(),ofstream::out | ofstream::app);
  out_<<gen_file<<","<<maxv_<<","<<maxc_<<","<<K_<<","<<num_scc<<","<<num_collapsed_nodes<<","<<num_iterations_<<",";
  out_<<adr_constr<<","<<cp_constr<<","<<ld_constr<<","<<st_constr<<",";
  out_<<fixed<<setprecision(2);
  out_<<load_time<<","<<store_time<<","<<ins_time<<","<<del_time<<","<<proc_time<<","<<analy_time<<","<<prog_time<<",";
  out_<<tpts<<","<<diffpts<<","<<get_mem_usage()<<"KB\n";
  /*  out_<<gen_file<<","<<maxv_<<","<<maxc_<<","<<K_<<","<<num_scc<<","<<num_collapsed_nodes<<","<<num_iterations_<<",";
   *  out_<<analy_time<<","<<tpts<<","<<diffpts<<","<<get_mem_usage()<<"KB\n";
   */
  out_.close();
}

void KitSolver::printLevel() {
  cout<<"Level:"<<endl;
}

void KitSolver::printEdges() {
  cout<<"New Edges"<<endl;
  for(list<NodePair>::iterator it=newEdges.begin();it!=newEdges.end();it++){
    cout<<(*it).u_<<"_"<<(*it).v_<<"\t";
  }
  cout<<endl;
}

void KitSolver::init() {
  u32 n, f, p;
  std::ifstream in(gen_file.c_str());
  bitmap_obstack_initialize(NULL);
  /* Variable n is the number of function parameters to be read.  */
  in >> maxv_ >> f >> n;
  minv_ = 0;
  /* Create the constraint graph.  */
  node_list_.insert(node_list_.begin(), maxv_, Node());
  for (u32 i = 0; i < maxv_; i++) {
    node_list_[i].init();
  }
  /* create the map from function --> # parameters.  */
  for (u32 i = 0; i < n; i++) {
    in >> f >> p;
    fp_[f] = p;
  }
  /* Read the constraints:  */
  in >> maxc_;
  cp_constr=0;
  adr_constr=0;
  st_constr=0;
  ld_constr=0;
  /* fill in the initial constraint graph and points-to sets, and the
   *    list of complex constraints. Each line of the input file is in the
   *    format lhs_var_name offset is_deref is_address rhs_var_name offset
   *    is_deref is_address.  */
  for (u32 i = 0; i < maxc_; i++) {
    u32 lv, lk, ld, la, rv, rk, rd, ra;
    in >> lv >> lk >> ld >> la;
    in >> rv >> rk >> rd >> ra;
    AccessPath l(lv, lk, ld, la);
    AccessPath r(rv, rk, rd, ra);
    assert(l.v_ < maxv_ && r.v_ < maxv_ && !l.adr_ && !(l.drf_ && r.drf_)
    && !(r.drf_ && r.adr_) && !(l.drf_ && r.adr_));
    if (r.adr_) {
      adr_constr++;
      bitmap_set_bit(node_list_[l.v_].pts_, r.v_);
    } 
    else if (r.drf_ || l.drf_) {
      if(r.drf_)
	ld_constr++;
      else
	st_constr++;
      constraints_.push_back(Constraint(l, r));
    } 
    else {
      bitmap_set_bit(node_list_[r.v_].edges_, l.v_);
      cp_constr++;
      newEdges.push_back(NodePair(r.v_,l.v_));
    }
  }
  // collapse pointer-equivalent variables.  
  if (in >> n) {
    u32 v, x;
    std::string line;
    getline(in,line);
    for (u32 i = 0; i < n; i++) {
      getline(in,line);
      std::istringstream inl(line);
      inl >> v;
      while (inl >> x) { unify_nodes(x, v); }
    }
  }
  changedlist= BITMAP_ALLOC(NULL);
  counter=0;
  levellist=new map<u64,NodePair>[K_];
  ins_time=0;
  del_time=0;
  proc_time=0;
  analy_time=0;
  load_time=0;
  store_time=0;
  in.close();
}

int main (int argc, char **argv) {  
  if (argc != 5) {
    cerr << "Syntax: " << argv[0] << " K file.gen ptsto stats\n";
    return 1;
  } 
  else {
    double start_time,end_time,start_time2,end_time2;
    start_time=get_time();
    KitSolver ks_obj;     
    ks_obj.K_=atoi(argv[1]);
    cerr<<"running for K= "<<ks_obj.K_<<"....."<<endl;
    ks_obj.gen_file=argv[2];
    ks_obj.ptsto_file=argv[3];
    ks_obj.stats_file=argv[4];
    
    ks_obj.init();
    start_time2 = get_time();
    ks_obj.solve();
    end_time2 = get_time();
    ks_obj.analy_time=end_time2-start_time2;
    
    end_time=get_time();
    ks_obj.prog_time=end_time-start_time;
    //ks_obj.print_ptsto();
    ks_obj.print_statistics();
  }
}
